#include <QtGui>
#include <QtOpenGL>
#include <GL/glut.h>

#include "GlWidget.h"

#ifndef GL_MULTISAMPLE
#define GL_MULTISAMPLE  0x809D
#endif

GLWidget::GLWidget(QWidget *parent)
    : QGLWidget(QGLFormat(QGL::SampleBuffers), parent)
{
    xRot = 0;
    yRot = 0;
    zRot = 0;
    _eyeZ = 500;

    xInt.slew( 0.3 );
    yInt.slew( 0.3 );
}

GLWidget::~GLWidget()
{
}

// @Override
QSize GLWidget::minimumSizeHint() const
{
    return QSize(50, 50);
}

// @Override
QSize GLWidget::sizeHint() const
{
    return QSize(800, 600);
}

/**
 * Adds a graphics object.
 */
void GLWidget::addGfxObject( GfxObject* o )
{
    _objects.push_back( o );
}

/**
 * The timer callback.
 */
void GLWidget::animate()
{
    repaint();
}

static void qNormalizeAngle(int &angle)
{
    while (angle < 0)
        angle += 360;
    while (angle > 360)
        angle -= 360;
}

void GLWidget::setXRotation(int angle)
{
//    qNormalizeAngle(angle);
    if (angle != xRot) {
        xRot = angle;
        xInt.goal(xRot);
//        emit xRotationChanged(angle);
        updateGL();
    }
}

void GLWidget::setYRotation(int angle)
{
//    qNormalizeAngle(angle);
    if (angle != yRot) {
        yRot = angle;
        yInt.goal(yRot);
//        emit yRotationChanged(angle);
        updateGL();
    }
}

void GLWidget::setZRotation(int angle)
{
    qNormalizeAngle(angle);
    if (angle != zRot) {
        zRot = angle;
//        emit zRotationChanged(angle);
        updateGL();
    }
}

/**
 * Initializes OpenGL
 */
void GLWidget::initializeGL()
{
    glClearColor(0, 0, 0, 0);

    glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);	// The Type Of Depth Test To Do
    glShadeModel(GL_SMOOTH);
    glEnable(GL_MULTISAMPLE);    
    glEnable(GL_POLYGON_SMOOTH);
	glHint (GL_POLYGON_SMOOTH_HINT, GL_NICEST);

    // alpha
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // material
    glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
    
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    
    static GLfloat lightPosition[4] = { 200, 300, 500, 1.0 };
    glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);

    // attenuation parameters
    glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1);
    glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.001);
    glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0);
}

/**
 * The Draw function
 */
void GLWidget::paintGL()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
     // create the viewing frustum    
    gfxPerspective( _width / _height, 160, 1, 1000, 500 );
    gluLookAt( 0.0f, 0.0f, _eyeZ, 
              0.0f, 0.0f, 0.0f, 
              0.0f, 1.0f, 0.0f );
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    // draw objects
    for( unsigned int i = 0; i < _objects.size(); i++ )
    {
        _objects.at( i )->drawObject();

        if( i == 0 ) 
        { // rotate the pad controller. this is a dirty quick hack
            xInt.interpolate();
            yInt.interpolate();
            _objects.at(i)->rotate( xInt.value(), yInt.value(), zRot );
        }
    }
}

/**
 * The Resize function
 */
void GLWidget::resizeGL(int width, int height)
{
    glViewport( 0, 0, width, height );

    _width = width;
    _height = height;

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
    // create the viewing frustum
    GLfloat ratio = (GLfloat) width / (GLfloat) height;
//    gfxOrtho( ratio, 16, 1, 100 );
    gfxPerspective( ratio, 160, 1, 1000, 500 );

    gluLookAt( 0.0f, 0.0f, _eyeZ, 
              0.0f, 0.0f, 0.0f, 
              0.0f, 1.0f, 0.0f );
}

// @Override
void GLWidget::mousePressEvent(QMouseEvent *event)
{
    _lastPos = event->pos();
}

// @Override
void GLWidget::mouseDoubleClickEvent( QMouseEvent *event )
{
    _lastPos = event->pos();
    
    setXRotation( 0 );
    setYRotation( 0 );
}

// @Override
void GLWidget::mouseMoveEvent(QMouseEvent *event)
{
    int dx = event->x() - _lastPos.x();
    int dy = event->y() - _lastPos.y();

    if (event->buttons() & Qt::LeftButton) 
    {
        setYRotation(yRot + 4 * dy);
    } 
    else if (event->buttons() & Qt::RightButton) 
    {
        setXRotation(xRot + 4 * dx);
    }
    
    _lastPos = event->pos();
}

/**
 * Interfaces glOrtho()
 */
void GLWidget::gfxOrtho( float ratio, float height, float near, float far )
{
    glOrtho(-height/2 * ratio, height/2 * ratio, -height/2, height/2, near, 
            far);
}

/**
 * Interfaces gluPerspective().
 */
void GLWidget::gfxPerspective( float ratio, float height, float near, float far, 
                              float depth )
{
    float angle = 2*::atan( height/2 / depth ) * 180 / PI;

    gluPerspective( angle, ratio, near, far );
}