#include <QtGui>

#include "Window.h"

#include "KeyState.h"

#include "tinyxml.h"
#include "XmlPrinter.h"

/**
 * Constructor
 */
Window::Window()
: QMainWindow()
,_padController(NULL)
,_chordSelector(NULL)
,_padPlot(NULL)
,_chordSuggestor(NULL)
,_suggestLabel(NULL)
{
    setWindowTitle( tr("Chord Collider") );
    setWindowIcon( QIcon( "res/pads.svg" ) );

    setCentralWidget( &_glWidget );

    QTimer *timer = new QTimer( this );
    connect( timer, SIGNAL( timeout() ), &_glWidget, SLOT( animate() ) );
    timer->start( 50 );

    createMenus();
}

/**
 * Creates menus.
 */
void Window::createMenus()
{    
    QMenu* fileMenu = menuBar()->addMenu( tr("&File") );
    
    QAction* openAct = new QAction( tr("&Open..."), this );
    openAct->setShortcuts( QKeySequence::Open );
    openAct->setStatusTip( tr( "Open an existing file" ) );
    connect( openAct, SIGNAL( triggered() ), this, SLOT( openPadDataFile() ) );
    fileMenu->addAction( openAct );
    
    QAction* saveAct = new QAction( tr("&Save..."), this );
    saveAct->setShortcuts( QKeySequence::Save );
    saveAct->setStatusTip( tr( "Save data to file" ) );
    connect( saveAct, SIGNAL( triggered() ), this, SLOT( savePadToFile() ) );
    fileMenu->addAction( saveAct );

    fileMenu->addSeparator();
    
    QAction* quitAct = new QAction( tr("&Quit..."), this );
    quitAct->setShortcuts( QKeySequence::Quit );
    quitAct->setStatusTip( tr( "Quits the application" ) );
    connect( quitAct, SIGNAL( triggered() ), this, SLOT( close() ) );
    fileMenu->addAction( quitAct );
}

/**
 * Handle adding chord to pad.
 */
void Window::handleAddChord()
{
    Chord* c = _chordSelector->currentChord();
    if( _padController != NULL ) _padController->addChord( *c );
}

/**
 * Dumps a chord column
 */
void Window::dumpChordColumn()
{
    ChordColumn* c = _chordSelector->currentColumn();
    if( _padController != NULL ) 
    {
        for( unsigned int i = 0; i < ITEMS_PER_COL; i++ )
        {
            _padController->addChord( c->getChord( i ) );
        }
    }
}

/**
 * Emulates pressing a pad.
 */
void Window::pressPad( unsigned int i )
{
    if( _padController != NULL ) 
    {
        MidiEvent me( 144, 0, 100, 0 );
        _padController->midiRemapper().sendPadMidi( i, me );
    }
}

/**
 * Emulates releasing a pad.
 */
void Window::releasePad( unsigned int i )
{ 
    if( _padController != NULL ) 
    {
        MidiEvent me( 128, 0, 100, 0 ); // TODO: use mouse position for 
        _padController->midiRemapper().sendPadMidi( i, me );
    }
}

/**
 * Clears a pad.
 */
void Window::clearPad( unsigned int i )
{
    if( _padController != NULL ) _padController->clearPad( i );
}

/**
 * Sets the play mode.
 */
void Window::setPlayMode( bool b )
{    
    // handle changes
    if( b )
    {
        if( _chordSelector != NULL ) _chordSelector->hide();
        if( _padPlot != NULL ) _padPlot->show();
    }
    else
    {
        if( _chordSelector != NULL ) _chordSelector->show();
        if( _padPlot != NULL ) _padPlot->hide();
    }
    KeyState::INSTANCE()->playMode( b );
}

/**
 * Saves the pad settings in an XML file.
 */
void Window::savePadToFile()
{
    if( _padController == NULL ) return; // TODO: throw warning

    // show file dialog to select file name.
    QString fileName = QFileDialog::getSaveFileName( 
                                        this, tr("Save Pad Data File"), "./",
                                        tr("XML Files (*.xml)") );
    if( fileName.isNull() ) return;

    // create xml document
    TiXmlDocument doc;  
    TiXmlDeclaration* decl = new TiXmlDeclaration( "1.0", "", "" );
    doc.LinkEndChild( decl );

    TiXmlElement* root = new TiXmlElement( "PadController" );
    doc.LinkEndChild( root );

    // save pad datas
    MidiRemapper& midiMap = _padController->midiRemapper();

    for( unsigned int i = 0; i < midiMap.numPads(); i++ )
    {
        PadData& pd = midiMap.padData( i );
        TiXmlElement* pad = new TiXmlElement( "Pad" );
        root->LinkEndChild( pad );
        
        pad->SetAttribute( "id", pd.id() );
        pad->SetAttribute( "inputNote", pd.inputNote() );

        Chord& c = pd.chord();
        TiXmlElement* chord = new TiXmlElement( "Chord" );
        pad->LinkEndChild( chord );

        chord->SetAttribute( "octave", c.octave() );
        chord->SetAttribute( "label", c.label().c_str() );

        vector<int>* sv = c.scaleValues();
        TiXmlElement* scaleValues = new TiXmlElement( "ScaleValues" );
        chord->LinkEndChild( scaleValues );

        for( unsigned int n = 0; n < c.noteCount(); n++ )
        {
            TiXmlElement* scaleValue = new TiXmlElement( "ScaleValue" );
            scaleValues->LinkEndChild( scaleValue );
            
            scaleValue->SetAttribute( "note", sv->at( n ) );
        }
    }

    // save document to file
    if( !fileName.endsWith( ".xml" ) ) fileName.append( ".xml" );
    doc.SaveFile( fileName.toStdString().c_str() );; 
}

/**
 * Opens pad data file and loads it to the pad controller.
 */
void Window::openPadDataFile()
{
     if( _padController == NULL ) return; // TODO: throw warning

    // open file.
    QString fileName = QFileDialog::getOpenFileName( 
                                        this, tr("Open Pad Data File"), "./",
                                        tr("XML Files (*.xml)") );
    if( fileName.isNull() ) return;

    // load file to document
    TiXmlDocument doc( fileName.toStdString().c_str() );
	bool loadOkay = doc.LoadFile();
	if( !loadOkay ) 
    {
        cout << "Failed to load file: " << fileName.toStdString() << endl;
        return;
    }
    
    // load data to pad controller
    MidiRemapper& midiMap = _padController->midiRemapper();
    
    TiXmlNode* padCon;
    TiXmlAttribute* attr;
    for( padCon = doc.FirstChild(); padCon != NULL; padCon = padCon->NextSibling() )
	{
        if( padCon->Type() != TiXmlNode::TINYXML_ELEMENT ) continue;
        
        TiXmlNode* pad;
        for( pad = padCon->FirstChild(); pad != NULL; pad = pad->NextSibling() )
        { // add pad data
            if( pad->Type() != TiXmlNode::TINYXML_ELEMENT ) continue;
            TiXmlElement* pe = pad->ToElement();
            
            // get pad id
            int id = 0;
            attr = pe->FirstAttribute();
            attr->QueryIntValue( &id );

            PadData& pd = midiMap.padData( (unsigned int) id );

            // get inputNote
            int inputNote = 0;
            attr = attr->Next();
            attr->QueryIntValue( &inputNote );

            pd.inputNote( (unsigned int) inputNote );
            
            // get chord data
            Chord& c = pd.chord();
            c.clear();
            
            TiXmlNode* chord = pad->FirstChild();
            if( chord->Type() != TiXmlNode::TINYXML_ELEMENT ) continue;
            TiXmlElement* ce = chord->ToElement();
            
            // get octave
            int octave = 0;
            attr = ce->FirstAttribute();
            attr->QueryIntValue( &octave );

            c.octave( octave );

            // get label
            attr = attr->Next();
            string label( attr->Value() );
            c.label( label );

            // get scale values
            TiXmlNode* sv = chord->FirstChild();
            TiXmlNode* note;
            for( note = sv->FirstChild(); note != NULL; note = note->NextSibling() )
            {
                if( note->Type() != TiXmlNode::TINYXML_ELEMENT ) continue;
                TiXmlElement* ne = note->ToElement();

                int nv = 0;
                attr = ne->FirstAttribute();
                attr->QueryIntValue( &nv );

                c.addScaleValue( (unsigned int) nv );
            }

            // update pad data.
            pd.update();
        }
	}

    _padController->updateUI();

}

// @Override
void Window::keyReleaseEvent( QKeyEvent *e )
{
    if( e->key() == Qt::Key_Return || e->key() == Qt::Key_Enter )
    { // stop chord playing
        if( e->isAutoRepeat() ) return; // don't handle auto repeat.
        if( _chordSelector != NULL 
           && _chordSelector->isVisible() ) _chordSelector->stop();
    }
    // keyboard pads!
    else if( e->key() == Qt::Key_C && !e->isAutoRepeat() ) releasePad( 0 );
    else if( e->key() == Qt::Key_V && !e->isAutoRepeat() ) releasePad( 1 );
    else if( e->key() == Qt::Key_B && !e->isAutoRepeat() ) releasePad( 2 );
    else if( e->key() == Qt::Key_N && !e->isAutoRepeat() ) releasePad( 3 );
    else if( e->key() == Qt::Key_D && !e->isAutoRepeat() ) releasePad( 4 );
    else if( e->key() == Qt::Key_F && !e->isAutoRepeat() ) releasePad( 5 );
    else if( e->key() == Qt::Key_G && !e->isAutoRepeat() ) releasePad( 6 );
    else if( e->key() == Qt::Key_H && !e->isAutoRepeat() ) releasePad( 7 );
    else if( e->key() == Qt::Key_E && !e->isAutoRepeat() ) releasePad( 8 );
    else if( e->key() == Qt::Key_R && !e->isAutoRepeat() ) releasePad( 9 );
    else if( e->key() == Qt::Key_T && !e->isAutoRepeat() ) releasePad( 10 );
    else if( e->key() == Qt::Key_Y && !e->isAutoRepeat() ) releasePad( 11 );
    else if( e->key() == Qt::Key_3 && !e->isAutoRepeat() ) releasePad( 12 );
    else if( e->key() == Qt::Key_4 && !e->isAutoRepeat() ) releasePad( 13 );
    else if( e->key() == Qt::Key_5 && !e->isAutoRepeat() ) releasePad( 14 );
    else if( e->key() == Qt::Key_6 && !e->isAutoRepeat() ) releasePad( 15 );
    
    QWidget::keyReleaseEvent(e);
}

// @Override
void Window::keyPressEvent( QKeyEvent *e )
{
    if( e->key() == Qt::Key_Plus || e->key() == Qt::Key_Equal )
    {
        KeyState* ks = KeyState::INSTANCE();
        int key = ks->key();
        key++;
        if( key > 11 ) key -= 12;
        ks->key( key );
    }
    else if( e->key() == Qt::Key_Minus || e->key() == Qt::Key_Underscore )
    {
        KeyState* ks = KeyState::INSTANCE();
        int key = ks->key();
        key--;
        if( key < 0 ) key += 12;
        ks->key( key );
    }
    else if( e->key() == Qt::Key_BracketLeft )
    { // change the scale mode
        KeyState* ks = KeyState::INSTANCE();
        ks->mode( ModeFactory::INSTANCE()->prevMode( ks->mode() ) );
    }
    else if( e->key() == Qt::Key_BracketRight )
    { // change the scale mode
        KeyState* ks = KeyState::INSTANCE();
        ks->mode( ModeFactory::INSTANCE()->nextMode( ks->mode() ) );
    }
    else if( e->key() == Qt::Key_Up )
    { // select previous chord
        if( _chordSelector != NULL  
               && _chordSelector->isVisible() ) _chordSelector->prev();
    }
    else if( e->key() == Qt::Key_Down )
    { // select next chord
        if( _chordSelector != NULL  
               && _chordSelector->isVisible() ) _chordSelector->next();
    }
    else if( e->key() == Qt::Key_Left )
    { // select previous chord column
        if( _chordSelector != NULL 
            && _chordSelector->isVisible() ) _chordSelector->prevColumn();
    }
    else if( e->key() == Qt::Key_Right )
    { // select next chord column
        if( _chordSelector != NULL  
            && _chordSelector->isVisible() ) _chordSelector->nextColumn();
    }
    else if( e->key() == Qt::Key_Return || e->key() == Qt::Key_Enter )
    { // play chord
        if( e->modifiers() == Qt::ControlModifier ) 
        {
            if( _chordSelector != NULL 
               && _chordSelector->isVisible() ) handleAddChord();
        }
        else
        {
            if( e->isAutoRepeat() ) return; // don't play on auto repeat.
            if( _chordSelector != NULL 
               && _chordSelector->isVisible() ) _chordSelector->play();
        }
    }
    else if( e->key() == Qt::Key_Space )
    { // toggle play mode
        setPlayMode( !KeyState::INSTANCE()->isPlayMode() );
    }
    else if( e->key() == Qt::Key_Delete )
    { // clear pads
        if( _padController != NULL ) _padController->clearAll();
    }
    else if( e->key() == Qt::Key_Insert )
    { // dump column
        if( _chordSelector != NULL 
               && _chordSelector->isVisible() ) dumpChordColumn();
    }
    else if( e->key() == Qt::Key_Comma ) 
    { // decrease octave
        if( _chordSelector != NULL 
               && _chordSelector->isVisible() ) _chordSelector->decOctave();
    }
    else if( e->key() == Qt::Key_Period ) 
    { // increase octave
        if( _chordSelector != NULL 
               && _chordSelector->isVisible() ) _chordSelector->incOctave();
    }
    else if( e->key() == Qt::Key_0 ) 
    { 
        if( e->modifiers() == Qt::ControlModifier ) 
        { // reset stats
            if( _padPlot != NULL ) _padPlot->reset();
        }
    }
    else if( e->key() == Qt::Key_P ) 
    { // toggle stats plot depth
        if( _padPlot != NULL && _padPlot->isVisible() ) 
            _padPlot->enableDepth( !_padPlot->isDepthEnabled() );
    }
    else if( e->key() == Qt::Key_M ) 
    { // toggle suggest mode
        if( _chordSuggestor != NULL )
        {
            _chordSuggestor->useStats( !_chordSuggestor->isUsingStats() );

            if( _suggestLabel == NULL ) return;
            
            string l( "Suggest Mode: " );
            if( _chordSuggestor->isUsingStats() ) l.append( "Statistical" );
            else l.append( "Tonal" );
            
            _suggestLabel->setText( l );
        }
    }
    // keyboard pads!
    else if( e->key() == Qt::Key_C && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 0 );
        else pressPad( 0 );
    }
    else if( e->key() == Qt::Key_V && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 1 );
        else pressPad( 1 );
    }
    else if( e->key() == Qt::Key_B && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 2 );
        else pressPad( 2 );
    }
    else if( e->key() == Qt::Key_N && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 3 );
        else pressPad( 3 );
    }
    else if( e->key() == Qt::Key_D && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 4 );
        else pressPad( 4 );
    }
    else if( e->key() == Qt::Key_F && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 5 );
        else pressPad( 5 );
    }
    else if( e->key() == Qt::Key_G && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 6 );
        else pressPad( 6 );
    }
    else if( e->key() == Qt::Key_H && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 7 );
        else pressPad( 7 );
    }
    else if( e->key() == Qt::Key_E && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 8 );
        else pressPad( 8 );
    }
    else if( e->key() == Qt::Key_R && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 9 );
        else pressPad( 9 );
    }
    else if( e->key() == Qt::Key_T && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 10 );
        else pressPad( 10 );
    }
    else if( e->key() == Qt::Key_Y && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 11 );
        else pressPad( 11 );
    }
    else if( e->key() == Qt::Key_3 && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 12 );
        else pressPad( 12 );
    }
    else if( e->key() == Qt::Key_4 && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 13 );
        else pressPad( 13 );
    }
    else if( e->key() == Qt::Key_5 && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 14 );
        else pressPad( 14 );
    }
    else if( e->key() == Qt::Key_6 && !e->isAutoRepeat() ) 
    {
        if( e->modifiers() == Qt::ControlModifier ) clearPad( 15 );
        else pressPad( 15 );
    }
    
    QWidget::keyPressEvent(e);
}