#include "GfxObject.h"

/**
 * Constructor
 */
GfxObject::GfxObject()
:_isVisible(true)
,_color(255, 255, 255) // default white
,_emission(0)
,_specular(10)
,_reflectance(0.5)
,_metallic(false)
{
}

/**
 * Destructor
 */
GfxObject::~GfxObject()
{
}

/**
 * Sets the translation.
 */
void GfxObject::translate( float x, float y, float z )
{
    _translation.setX( x );
    _translation.setY( y );
    _translation.setZ( z );
}

/**
 * Sets the rotation. 
 * angles in degrees x -> yaw, y -> tilt, z -> roll
 */
void GfxObject::rotate( float x, float y, float z )
{
    _rotation.setX( x );
    _rotation.setY( y );
    _rotation.setZ( z );
}

/**
 * Sets the color.
 */
void GfxObject::color( int r, int g, int b, int a )
{
    // TODO: validate values
    
    _color.setRgb( r, g, b, a );
}

/**
 * Sets the color
 */
void GfxObject::color( QColor& c )
{
    _color = c;
}

/**
 * updates the color and material values.
 */
void GfxObject::updateMaterial()
{
    // set color
    // it is possible to set per vertex color 
    // in subclass' drawComponent() method.
    glColor4f( _color.redF(), _color.greenF(), _color.blueF(), 
              _color.alphaF() );

    // set material values
    // set ambient and diffuse color
    float c[3];
    c[0] = _color.redF();
    c[1] = _color.greenF();
    c[2] = _color.blueF();
    glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, c );

    // specular color
    float specColor[3] = {_reflectance, _reflectance, _reflectance};
    if( _metallic ) 
    {   
        specColor[0] *= c[0];
        specColor[1] *= c[1];
        specColor[2] *= c[2];
    } 
    glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, specColor );

    // set shininess
    glMaterialf( GL_FRONT_AND_BACK, GL_SHININESS, _specular );

    float em[3] = {0, 0, 0};
    if( _emission > 0 ) 
    {
        em[0] = c[0] * _emission;
        em[1] = c[1] * _emission;
        em[2] = c[2] * _emission;
    } 
    glMaterialfv( GL_FRONT_AND_BACK, GL_EMISSION, em );
}

/**
 * Draws the graphics object. General operations such as translation, 
 * rotation, material setting is done here.
 * The actual rendering is handled in drawComponent(). 
 */
void GfxObject::drawObject()
{
    if( !_isVisible ) return;
    
    // save transformations
	glPushMatrix();
	
	// translate object
	glTranslatef( _translation.x(), _translation.y(), _translation.z() );
	
	// rotate object 
	glRotatef( _rotation.x(), 0, 1, 0 ); // yaw
	glRotatef( _rotation.y(), 1, 0, 0 ); // tilt
	glRotatef( _rotation.z(), 0, 0, 1 ); // roll

    // update colors and material
    updateMaterial();

	// draw the component. this is handled in the subclasses.
	drawComponent();
	
	// reset transformations
	glPopMatrix();
}