#include <QApplication>
#include <QDesktopWidget>
#include <QSplashScreen>
#include <GL/glut.h>

#include "Window.h"
#include "MidiEngine.h"

#include "KeyState.h"
#include "PadController.h"
#include "ChordSelector.h"
#include "KeyStateLabel.h"
#include "ChordSuggestor.h"
#include "PadStatistics.h"
#include "PadStatBoxPlot.h"
//#include "TextSuggestor.h"

void showMessage( QSplashScreen& splash, const char* message )
{
    splash.showMessage( message, Qt::AlignBottom | Qt::AlignLeft, 
                       QColor( 255,255,255 ) );
    usleep( 500000 ); // wait to show splash
}

/**
 * The entry point
 */
int main(int argc, char *argv[])
{
    QApplication app( argc, argv );

    QPixmap pixmap( "res/collider.png" );
    QSplashScreen splash(pixmap);
    splash.show();

    // initialize GLUT
    showMessage( splash, "Initializing GLUT" );
    glutInit( &argc, argv );

    // Initialize MidiEngine
    showMessage( splash, "Initializing MIDI Engine" );
    MidiEngine midiEngine;
    midiEngine.initialize();
    
    // Initialize Window
    showMessage( splash, "Initializing UI" );
    Window window;
    window.resize( window.sizeHint() );

    // Initialize key state
    showMessage( splash, "Initializing UI Components" );
    KeyState* keyState = KeyState::INSTANCE();
    keyState->key( 0 ); // C

    // Create and initialize objects. 
    // TODO: seg-faults when moved to a function.

    // Create Pad Controller
    showMessage( splash, "Initializing PadController" );
    PadController padController( midiEngine );
    window.glWidget().addGfxObject( &padController.pcObject() );
    window.padController( &padController );
    KeyState::INSTANCE()->addListener( &padController ); 

    // Create Chord Columns
    showMessage( splash, "Initializing ChordSelector" );
    ChordSelector chordSelector( midiEngine );
    chordSelector.translate( 80, 0, 0 );
    window.glWidget().addGfxObject( &chordSelector );
    window.chordSelector( &chordSelector );
    
    // Create and add labels
    showMessage( splash, "Initializing Labels" );
    KeyStateLabel ksLabel;
    ksLabel.translate( -55, 65, 0 );
    window.glWidget().addGfxObject( &ksLabel );
    ksLabel.updateText();
    KeyState::INSTANCE()->addListener( &ksLabel );

    // Add pad statistics
    showMessage( splash, "Initializing Statistics" );
    PadStatistics padStats;
	padController.midiRemapper().addPadEventListener( &padStats );

    // Add pad statistics plot
    PadStatBoxPlot psBoxPlot( padStats );
    psBoxPlot.translate( -95, -40, 0 );
    window.padPlot( &psBoxPlot );
    window.glWidget().addGfxObject( &psBoxPlot );
    
    // Create chord suggestor
    showMessage( splash, "Initializing ChordSuggestor" );
    ChordSuggestor chordSuggestor( padController.midiRemapper(), 
                                  padController.pcObject(),
                                  chordSelector, padStats );
	padController.midiRemapper().addPadEventListener( &chordSuggestor );
    window.chordSuggestor( &chordSuggestor );

    string l( "Suggest Mode: Tonal" );
    GfxFtStringObject suggestLabel( l );
    suggestLabel.setFontSize( 7 );
    suggestLabel.translate( -55, -70, 0 );
    window.glWidget().addGfxObject( &suggestLabel );
    window.suggestLabel( &suggestLabel );

    // Add text suggestor
//    TextSuggestor textSuggestor( padStats );
//    textSuggestor.translate( -55, -75, 0 );
//    window.glWidget().addGfxObject( &textSuggestor );

    window.setPlayMode( false );
    window.show();     
    
    splash.finish(&window);
    return app.exec();
}