#include "MidiEngine.h"

#include <vector>

/**
 * Default constructor
 */
MidiEngine::MidiEngine()
:_rtMidiIn(string("HxMidiIn"))
,_rtMidiOut(string("HxMidiOut"))
{
}

/**
 * Destructor
 */
MidiEngine::~MidiEngine()
{
}

/**
 * Initializes the MIDI engine.
 */
void MidiEngine::initialize()
{
	// Check available input ports
	unsigned int nPorts = _rtMidiIn.getPortCount();
	cout << "Found " << nPorts << " MIDI input ports." << endl;

    unsigned int i = 0;
    int portNo = 0;
    string portName;
    // Open input port.
	try 
	{
        // search for MPD
        for ( i = 0; i < nPorts; i++ ) 
        {
            portName = _rtMidiIn.getPortName( i );
            if( portName.find( "MPD" ) != string::npos )
            {
                portNo = i;
                cout << "Using MPD on port " << portNo << endl;
                break;
            }
        }

        if( i == nPorts )
        {
            cout << "Could not find MPD. Using port 0 as default." << endl;
        }
        
		_rtMidiIn.openPort( portNo );
	}
	catch ( RtError &error ) 
	{
		error.printMessage();
	}

	// Set our callback function.  This should be done immediately after
	// opening the port to avoid having incoming messages written to the
	// queue instead of sent to the callback function.
	_rtMidiIn.setCallback( &callback , (void*) this );

	// Don't ignore sysex, timing, or active sensing messages.
	_rtMidiIn.ignoreTypes( false, false, false );

	// Check available output ports
    nPorts = _rtMidiOut.getPortCount();
	cout << "Found " << nPorts << " MIDI Output ports." << endl;
    
    // Open output port.
    portNo = 0;
    try
    {
        // search for FluidSynth
        for ( i = 0; i < nPorts; i++ ) 
        {
            portName = _rtMidiOut.getPortName( i );
            if( portName.find( "FLUID" ) != string::npos )
            {
                portNo = i;
                cout << "Using Fluid Synth on port " << portNo << endl;
                break;
            }
        }
        
        if( i == nPorts )
        {
            cout << "Could not find FluidSynth. Using port 0 as default." << endl;
        }
        
        _rtMidiOut.openPort( portNo ); // Using port 0 by default. Configure in Jack.
	}
	catch ( RtError &error ) 
	{
		error.printMessage();
	}
}

/**
 * Adds a listener.
 */
void MidiEngine::addListener( MidiEngineListener* l )
{
    _listeners.push_back( l );
}

/**
 * Sends a Midi message.
 */
void MidiEngine::sendMidiMessage( MidiEvent& e )
{
//    cout << "(MidiOut)";
//    e.print();

    _rtMidiOut.sendMessage( e.toVector() );
}

/** 
 * Handles the callback.
 */
void MidiEngine::handleCallback( double deltatime, 
                                vector< unsigned char > *message )
{
    int b0 = (int) message->at( 0 );
    int b1 = message->size() > 1 ? (int) message->at( 1 ) : -1;
    int b2 = message->size() > 2 ? (int) message->at( 2 ) : -1;

    MidiEvent e( b0, b1, b2, deltatime );
    
//    cout << "(MidiIn)";
//    e.print();

    for( unsigned int i = 0; i < _listeners.size(); i++ )
    {
        _listeners[i]->handleMidiMessage( e );
    }
}

/**
 * Callback wrapper
 */
void MidiEngine::callback( double deltatime, vector< unsigned char > *message, 
                          void *data )
{
	( (MidiEngine*) data )->handleCallback(deltatime, message);
}