#include "MidiRemapper.h"

#include "Common.h"

#include "KeyState.h"

/**
 * Constructor
 */
MidiRemapper::MidiRemapper( MidiEngine& e )
:_midiEngine(e)
{
}

/**
 * Destructor
 */
MidiRemapper::~MidiRemapper()
{
}

/**
 * Initializes the remapper. Temporarily, the pad data is created here.
 */
void MidiRemapper::initialize()
{
	initPadData();
}

/**
 * Initializes PadData objects.
 */
void MidiRemapper::initPadData()
{    
    unsigned int banks = 1;
    for( unsigned int b = 0; b < banks; b++ )
    {        
         for( unsigned int i = 0; i < 16; i++ )
        {
            // set inputs
            _padMap[i + b * 16 + 36] = PadData( i, b );
			_padIdList.push_back( i + b * 16 + 36 );

            PadData& pd = _padMap[i + b * 16 + 36];
            pd.inputNote( i + b * 16 + 36 ); // TODO: read from setup file.
        }
    }
}

/**
 * Sets the chord to the given pad.
 */
void MidiRemapper::setPadChord( unsigned int id, Chord& c )
{
    padData( id ).setChord( c );
}

/**
 * Adds a chord to the first empty slot.
 * Returns false if fails to find an empty slot.
 */
bool MidiRemapper::addChord( Chord& c, unsigned int* id )
{
    // find an empty pad.
    for( unsigned int i = 0; i < _padIdList.size(); i++ )
    {
        if( padData( i ).chord().noteCount() == 0 )
        { // empty?
            setPadChord( i, c );
            *id = i;
            return true;
        }
    }
    
    return false;
}

/**
 * Clears all pad data.
 */
void MidiRemapper::clearAllData()
{
     // find an empty pad.
    for( unsigned int i = 0; i < _padIdList.size(); i++ )
    {
        padData( i ).clear();
    }
}

/**
 * Updates the pad chords
 */
void MidiRemapper::updatePadChords()
{
    for( unsigned int i = 0; i < _padIdList.size(); i++ )
    {
        padData( i ).update();
    }
}

/**
 * Returns the PadData with the given ID.
 */
PadData& MidiRemapper::padData( unsigned int id )
{
	assert( id < _padIdList.size() );
	
	return _padMap[_padIdList[id]];
}

/**
 * Sends the midi message for the pad data.
 */
void MidiRemapper::sendPadMidi( unsigned int id, MidiEvent& e )
{
    PadData& pd = padData( id );
    unsigned int size = pd.outputNotes()->size();
       
    // fire pad event
    PadEvent pe( pd, e );
    firePadEvent( pe );
        
    // send MIDI output messages
    for( unsigned int i = 0; i < size; i++ )
    {
        MidiEvent m( e );
        m.note( pd.outputNotes()->at( i ) );
        
        _midiEngine.sendMidiMessage( m );
    }
}

/**
 * Adds a pad event listener. Usually used for UI objects.
 */
void MidiRemapper::addPadEventListener( PadEventListener* l )
{
    _listeners.push_back( l );
}

/**
 * @see MidiEngineListener::handleMidiMessage()
 */
void MidiRemapper::handleMidiMessage( MidiEvent& e )
{
    // only change note on/off messages. bypass other messages
    if( e.type() == 128 || e.type() == 144 ) 
    {
        map<int, PadData>::iterator iter;
        iter = _padMap.find( e.note() );
        if( iter == _padMap.end() ) return;

        PadData& pd = iter->second;

        // send MIDI output messages
        sendPadMidi( pd.id(), e );
    }
    else if( e.type() == 208 ) // modulation
    { // do nothing. pads are to sensitive for modulation.
    }
    else
    { // bypass message
        _midiEngine.sendMidiMessage( e );
    }
}

/**
 * Fires the pad event to all listeners.
 */
void MidiRemapper::firePadEvent( PadEvent& e )
{
    for( unsigned int i = 0; i < _listeners.size(); i++ )
    {
        _listeners[i]->handlePadEvent( e );
    }
}