#include "ChordColumn.h"

#include "KeyState.h"
#include "CommonUtil.h"

QColor PLAY_COLOR(192, 0, 0);
QColor IDLE_COLOR(0, 0, 192);

inline float calcYPos( int i )
{
    int n = ITEMS_PER_COL;
    int h = ChordItem::ITEM_HEIGHT;
    int spacing = 3;
    
    float val = -h / 2 - ( h + spacing ) * ( i - ( n - 1.0f ) / 2 );
    if( i == -1 ) val = val + 1;
    return val;
}

/**
 * Constructor
 */
ChordColumn::ChordColumn( MidiEngine& m )
:_midiEngine(m)
,_selectIdx(0)
,_isPlaying(false)
{
    initialize();
}

/**
 * Destructor
 */
ChordColumn::~ChordColumn()
{
}

/**
 * Sets the title.
 */
void ChordColumn::setTitle( string& s )
{
    _titleObject.setText( s );
}

/**
 * Sets the title.
 */
void ChordColumn::setTitle( const char* s )
{
     _titleObject.setText( s );
}

/**
 * Selects the next chord.
 */
void ChordColumn::next()
{
    unsigned int idx = ( _selectIdx == (ITEMS_PER_COL-1) ) ? 
        0 : _selectIdx + 1;
    selectChord( idx );
}

/**
 * Selects the previous chord.
 */
void ChordColumn::prev()
{
    unsigned int idx = (_selectIdx == 0 ) ? 
        (ITEMS_PER_COL-1) : _selectIdx - 1;
    selectChord( idx );
}

/**
 * Selects the chord item with the given index
 */
void ChordColumn::selectChord( unsigned int i )
{ 
    assert( i < ITEMS_PER_COL ); 

    stop();
    
    // move previous chord object back
    _chordObjects[_selectIdx].idle();
    
    _selectIdx = i; 

    // bring selected object forward
    _chordObjects[_selectIdx].select();

    // move select icon    
    _selectIcon.translate( ChordItem::ITEM_WIDTH / 2, calcYPos( i ), 0 );
}

/**
 * Initializes the object.
 */
void ChordColumn::initialize()
{   
    int w = ChordItem::ITEM_WIDTH;
    int h = ChordItem::ITEM_HEIGHT;

    // the title object
    _titleObject.setFontSize( 5 );
    _titleObject.emission( 1 );
    _titleObject.translate( -w / 2, calcYPos( -1 ), 0 );
    _objects.push_back( &_titleObject );

    // the select icon
    _selectIcon.width( 1 );
    _selectIcon.height( h );
    _selectIcon.depth( 3 );
    _selectIcon.color( IDLE_COLOR );
    _selectIcon.emission( 1 );
    _selectIcon.translate( w / 2, calcYPos( 0 ), 0 );
    _objects.push_back( &_selectIcon );

    // the chord objects
    for( unsigned int i = 0; i < ITEMS_PER_COL; i++ )
    {
        _chordObjects[i].translate( -w / 2, calcYPos( i ), 0 );
        _objects.push_back( &_chordObjects[i] );
    }
}

/**
 * Plays the current chord.
 */
void ChordColumn::play()
{
    KeyState* ks = KeyState::INSTANCE();
    
    vector<int>* notes = currentChord().scaleValues();
    unsigned int nc = currentChord().noteCount();
    unsigned int o = currentChord().octave();
    for( unsigned int n = 0; n < nc; n++ )
    {
        int note = ks->scaleToNote( notes->at( n ), o );
        if( ks->mode() == SMN_MINOR_H && notes->at(0) == 3
           && notes->at(n) == 7 ) note--;
        
        MidiEvent m( 144, note, 100, 0 );        
        _midiEngine.sendMidiMessage( m );
    }

    _selectIcon.color( PLAY_COLOR );
    
    _isPlaying = true;
}

/**
 * Stops playing the current chord.
 */
void ChordColumn::stop()
{
    if( !_isPlaying ) return;
    
    KeyState* ks = KeyState::INSTANCE();
    
    vector<int>* notes = currentChord().scaleValues();
    unsigned int nc = currentChord().noteCount();
    unsigned int o = currentChord().octave();
    for( unsigned int n = 0; n < nc; n++ )
    {
        int note = ks->scaleToNote( notes->at( n ), o );
        if( ks->mode() == SMN_MINOR_H && notes->at(0) == 3
           && notes->at(n) == 7 ) note--;
        
        MidiEvent m( 128, note, 0, 0 );        
        _midiEngine.sendMidiMessage( m );
    }
    
    _selectIcon.color( IDLE_COLOR );
    
    _isPlaying = false;
}

/**
 * Initializes the chords.
 *@param c the relative chord.
 */
void ChordColumn::setChords( Chord& c )
{
    vector<int>* sv = c.scaleValues();
    unsigned int nc = c.noteCount();
    for( int i = 0; i < ITEMS_PER_COL; i++ )
    {
        for( unsigned int n = 0; n < nc; n++ )
        {
            if( i == 7 ) _chords[i].addScaleValue( sv->at(n) + i - 7 );
            else _chords[i].addScaleValue( sv->at(n) + i );
        }
    }
}

/**
 * Set chord label.
 */
void ChordColumn::setChordLabel( unsigned int i, string& s )
{
    assert( i < ITEMS_PER_COL );
    _chordObjects[i].setText( s );
    _chords[i].label( s );
}

/**
 * Sets the octave of the chords
 */
void ChordColumn::setOctave( unsigned int octave )
{
    for( int i = 0; i < ITEMS_PER_COL; i++ )
    {
        if( i == 7 ) _chords[i].octave( octave+1 );
        else _chords[i].octave( octave );
    }
}

// @Override
void ChordColumn::drawComponent()
{
    for( unsigned int i = 0; i < _objects.size(); i++ )
    {
        _objects.at(i)->drawObject();
    }
}