#include "ChordSelector.h"

#include "CommonUtil.h"

string octaveString( "Octave: " );

/**
 * Constructor
 */
ChordSelector::ChordSelector( MidiEngine& m )
:_midiEngine(m)
,_currentIdx(0)
,_octave(4)
,_xInt(0, 0, 0.3f)
,_xPos(0)
,_outPos(50)
{
    initialize();
}

/**
 * Destructor
 */
ChordSelector::~ChordSelector()
{
}

/**
 * Creates chord columns
 */
void ChordSelector::initialize()
{
    // create octave label
    _octaveObject.setFontSize( 5 );
    string os( octaveString );
    os += ( _octave-1 + '1' );
    _octaveObject.setText( os );
    _octaveObject.translate( -ChordItem::ITEM_WIDTH / 2, -60, 0 );
    
    Chord c;
    
    // Notes
    c.clear();
    c.addScaleValue(1);
    addChordColumn( "Note", c, true );
    
    // Thirds
    c.clear();
    c.addScaleValue(1);
    c.addScaleValue(3);
    addChordColumn( "Thirds", c, true, "3" );
    
    // Power
    c.clear();
    c.addScaleValue(1);
    c.addScaleValue(5);
    addChordColumn( "Power", c, true, "5" );
    
    // Octaves
    c.clear();
    c.addScaleValue(1);
    c.addScaleValue(8);
    addChordColumn( "Octave", c, true, "8" );
    
    // Triads
    c.clear();
    c.addScaleValue(1);
    c.addScaleValue(3);
    c.addScaleValue(5);
    addChordColumn( "Triad", c, false );
    
    // Seventh
    c.clear();
    c.addScaleValue(1);
    c.addScaleValue(3);
    c.addScaleValue(5);
    c.addScaleValue(7);
    addChordColumn( "Seventh", c, false, " 7" );
    
    // Fourth
    c.clear();
    c.addScaleValue(1);
    c.addScaleValue(4);
    c.addScaleValue(5);
    addChordColumn( "Fourth", c, false, " 4" );
    
    // select triads
    selectColumn( 4 );
}

/**
 * Adds a chord column
 */
void ChordSelector::addChordColumn( const char* label, Chord& c, 
                                   bool isInterval, const char* append )
{
    ChordColumn* col = new ChordColumn( _midiEngine );
    
    col->setTitle( label );
    col->setVisible( false );
    col->setChords( c );
    col->setOctave( _octave );

    // set labels
    string l;
    for( unsigned int i = 0; i < ITEMS_PER_COL; i++ )
    {
        l.clear();
        l.append( CommonUtil::numToRoman( i % 7 + 1, isInterval ) );
        if( append != NULL ) l.append( append );
        
        col->setChordLabel( i, l );
    }

    _columns.push_back( col );

    // add selector indicator
    GfxBoxObject* dot = new GfxBoxObject( 2, 1, 1 );
    dot->translate( -ChordItem::ITEM_WIDTH / 2.0 
                   + ( 2 + 1 ) * ( _columns.size() - 1 ),
                   52,
                   0 );
    dot->color( ChordItem::IDLE_COLOR );
    _columnDots.push_back( dot );
}        

/**
 * Returns the current column
 */
ChordColumn* ChordSelector::currentColumn()
{
    if( _columns.size() == 0 ) return NULL;
    else return _columns.at( _currentIdx );
}

/**
 * Returns the current chord
 */
Chord* ChordSelector::currentChord()
{
    ChordColumn* col = currentColumn();
    if( col != NULL ) return &col->currentChord();
    else return NULL;
}

/**
 * Selects the next column
 */
void ChordSelector::nextColumn()
{
    unsigned int i = ( _currentIdx + 1 ) == _columns.size() ? 
        0 : _currentIdx + 1;

    selectColumn( i );
}

/**
 * Selects the next column
 */
void ChordSelector::prevColumn()
{
    unsigned int i = ( _currentIdx == 0 ) ? 
        ( _columns.size() - 1 ) : _currentIdx - 1;

    selectColumn( i );
}

/**
 * Select the given column
 */
void ChordSelector::selectColumn( unsigned int i )
{
    assert( i < _columns.size() );
    
    ChordColumn* col = currentColumn();
    if( col == NULL || _columns.size() == 1 ) return;

    // save row number
    unsigned int chordIdx = col->selectIdx();
    
    col->setVisible( false );
    col->stop();
    _columnDots.at( _currentIdx )->color( ChordItem::IDLE_COLOR );
    
    _currentIdx = i;

    col = currentColumn();
    col->selectChord( chordIdx );
    col->setVisible( true );
    _columnDots.at( _currentIdx )->color( ChordItem::SELECT_COLOR );
}

/**
 * Increases the octave.
 */
void ChordSelector::incOctave()
{
    if( _octave < 7 ) 
    {
        _octave++;
        updateColumns();
    }
}

/**
 * Decreases the octave.
 */
void ChordSelector::decOctave()
{
    if( _octave > 1 ) 
    {
        _octave--;
        updateColumns();
    }
}

// This method will break down if the the x position is changed via
// gfxObject.translate().setX(x);
// @Override
void ChordSelector::translate( float x, float y, float z )
{
    _xPos = x; // save the assigned position
    
    GfxObject::translate( x, y, z );
}

/**
 * Shows the chord selector.
 * TODO: a layout object should handle this.
 */
void ChordSelector::show()
{
    setVisible( true );
    _xInt.goal( 0 );
}

/**
 * Hides the chord selector.
 * TODO: a layout object should handle this.
 */
void ChordSelector::hide()
{
    _xInt.goal( _outPos );
}

/**
 * Updates the columns.
 */
void ChordSelector::updateColumns()
{   
    string os( octaveString );
    os += ( _octave-1 + '1' );
    _octaveObject.setText( os );
    
    for( unsigned int i = 0; i < _columns.size(); i++ )
    {
        _columns.at( i )->setOctave( _octave );
    }
}

void ChordSelector::next()
{
    ChordColumn* col = currentColumn();
    if( col != NULL ) col->next();
}

void ChordSelector::prev()
{
    ChordColumn* col = currentColumn();
    if( col != NULL ) col->prev();
}

void ChordSelector::play()
{
    ChordColumn* col = currentColumn();
    if( col != NULL ) col->play();
}

void ChordSelector::stop()
{
    ChordColumn* col = currentColumn();
    if( col != NULL ) col->stop();
}

// @Override
void ChordSelector::drawComponent()
{
    _octaveObject.drawObject();
    for( unsigned int i = 0; i < _columnDots.size(); i++ )
    {
        _columnDots.at(i)->drawObject();
    }
    
    for( unsigned int i = 0; i < _columns.size(); i++ )
    {
        _columns.at(i)->drawObject();
    }

    _xInt.interpolate();
    GfxObject::translate().setX( _xPos + _xInt.value() );
    if( ::abs(_xInt.value() - _outPos) < 1  ) 
    {
        if( isVisible() ) setVisible( false );
    }
}