#include "ChordSuggestor.h"

/**
 * Default constructor
 */
ChordSuggestor::ChordSuggestor( MidiRemapper& m, PadControllerObject& p, 
                               ChordSelector& c, PadStatistics& ps )
:_midiMap(m)
,_pcObject(p)
,_chordSelector(c)
,_padStats(ps)
{
    useStats( false );
}

/**
 * Destructor
 */
ChordSuggestor::~ChordSuggestor()
{
}

/**
 * Use the stats
 */
void ChordSuggestor::useStats( bool b )
{
    _useStats = b;
    
    _suggestMap = _useStats ? 
        (SuggestMap*)&_padStats : (SuggestMap*)&_tonalMap;
}

/**
 * Suggest tonal map
 */
void ChordSuggestor::suggestTonal( PadEvent& e )
{
    Chord& c = e.padData().chord();
    if( c.noteCount() == 0 ) return; // empty pad

    // TODO: do chord analysis here
    // assuming lowest note is the root(bass) note
    unsigned int root = c.root();
    
    vector<unsigned int>& sList = _suggestMap->getSuggestion( root - 1 );
    // returns possible triads
    
    for( unsigned int i = 0; i < sList.size(); i++ )
    {// light up suggestions here
        
        // create suggested triad
        Chord sc;
        sc.addScaleValue(0 + sList.at(i));
        sc.addScaleValue(2 + sList.at(i));
        sc.addScaleValue(4 + sList.at(i));
        
        // pad object
        lightPads( e, sc );

        // chord selector
        lightChordItems( e, sc );
    }
}

/**
 * Suggest stats map
 */
void ChordSuggestor::suggestStats( PadEvent& e )
{
    Chord& c = e.padData().chord();
    if( c.noteCount() == 0 ) return; // empty pad

    // TODO: do chord analysis here
    // assuming lowest note is the root(bass) note
    unsigned int root = c.root();
    
    vector<unsigned int>& sList = _suggestMap->getSuggestion( root - 1 );
    // returns probabilities

    unsigned int sum = 0;
    for( unsigned int i = 0; i < sList.size(); i++ )
    {
        sum += sList.at( i );
    }

    float mix = 0;
    for( unsigned int i = 0; i < sList.size(); i++ )
    {// light up suggestions here
        mix = sum == 0 ? 0 : sList.at( i ) / (float)sum;

        mix = 1 - ( 1 - mix ) * ( 1 - mix );
        // create suggested triad
        Chord sc;
        sc.addScaleValue(1 + i);
        sc.addScaleValue(3 + i);
        sc.addScaleValue(5 + i);

        // pad object
        lightPads( e, sc, mix );

        // chord selector
        lightChordItems( e, sc, mix );
    }
}

/**
 * Lights the pads
 */
void ChordSuggestor::lightPads( PadEvent& e, Chord& sc, float mix )
{// pad object
    
    unsigned int numPads = _pcObject.numPads();
    for( unsigned int n = 0; n < numPads; n++ )
    {
        // don't need to light pressed pad
        if( n == e.padData().id() ) continue; 
        
        PadData& pd = _midiMap.padData( n );
        Chord& pc = pd.chord();
        if( pc.noteCount() == 0 ) continue; // empty pad

        unsigned int common = sc.compare( pc );

        if( common == sc.noteCount() || common == pc.noteCount() )
        { // light up
            if( e.midiEvent().isNoteOn() ) 
            {
                _pcObject.pad( pd.id() ).suggest( mix 
                     * common / (float) pc.noteCount() );
            }
            else if( e.midiEvent().isNoteOff() ) 
            {
                if( !_pcObject.pad( pd.id() ).isHit() )
                    _pcObject.pad( pd.id() ).idle();
            }
        }
    } 
}

/**
 * Lights the chord items
 */
void ChordSuggestor::lightChordItems( PadEvent& e, Chord& sc, float mix )
{ 
    unsigned int colCount = _chordSelector.columnCount();
    for( unsigned int cn = 0; cn < colCount; cn++ )
    {// for each column
        ChordColumn* col = _chordSelector.getColumn( cn );

        for( unsigned int n = 0; n < ITEMS_PER_COL; n++ )
        { // for each chord items
            ChordItem& ci = col->getChordItem( n );
            Chord& cic = col->getChord( n );
            if( cic.noteCount() == 0 ) continue;
            
            unsigned int common = sc.compare( cic );
            
            if( common == sc.noteCount() || common == cic.noteCount() )
            { // light up
                if( e.midiEvent().isNoteOn() ) 
                {
                    ci.suggest( mix * common / (float) cic.noteCount() );
                }
                else if( e.midiEvent().isNoteOff() ) 
                {
                    ci.suggest( 0 );
                }
            }
        }
    } 
}

// @Override
void ChordSuggestor::handlePadEvent( PadEvent& e )
{
    if( _useStats ) suggestStats( e );
    else suggestTonal( e );
}